(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, SandstormDb, i, appName, addStaticAsset, verifiedEmails, unverifiedEmails, userPictureUrl, fetchPicture;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Users = new Mongo.Collection("users");                                                                    // 29
// The users collection is special and can be accessed through `Meteor.users`.                               // 30
// See https://docs.meteor.com/#/full/meteor_users.                                                          // 31
//                                                                                                           // 32
// There are two distinct types of entries in the users collection: identities and accounts. An              // 33
// identity contains personal profile information and typically includes some intrinsic method for           // 34
// authenticating as the owner of that information.                                                          // 35
//                                                                                                           // 36
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                 // 37
// Each account can have multiple identities linked to it. To log in as an account you must first            // 38
// authenticate as one of its linked identities.                                                             // 39
//                                                                                                           // 40
// Every user contains the following fields:                                                                 // 41
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally               // 42
//        stable SHA-256 ID of this identity, hex-encoded.                                                   // 43
//   createdAt: Date when this entry was added to the collection.                                            // 44
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 45
//   services: Object containing login data used by Meteor authentication services.                          // 46
//   expires: Date when this user should be deleted. Only present for demo users.                            // 47
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.              // 48
//                                                                                                           // 49
// Identity users additionally contain the following fields:                                                 // 50
//   profile: Object containing the data that will be shared with users and grains that come into            // 51
//            contact with this identity. Includes the following fields:                                     // 52
//       service: String containing the name of this identity's authentication method.                       // 53
//       name: String containing the chosen display name of the identity.                                    // 54
//       handle: String containing the identity's preferred handle.                                          // 55
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                // 56
//                an identicon will be used.                                                                 // 57
//       pronoun: One of "male", "female", "neutral", or "robot".                                            // 58
//   unverifiedEmail: If present, a string containing an email address specified by the user.                // 59
//                                                                                                           // 60
// Account users additionally contain the following fields:                                                  // 61
//   loginIdentities: Array of identity objects, each of which may include the following fields.             // 62
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                   // 63
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use            // 64
//                       a separate array here so that we can use a Mongo index to enforce the               // 65
//                       invariant that an identity only be a login identity for a single account.           // 66
//   primaryEmail: String containing this account's primary email address. Must be a verified adress         // 67
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()              // 68
//                 to do this checking automatically.                                                        // 69
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.        // 70
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 71
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 72
//               inviter admin attached to the key.                                                          // 73
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 74
//                the invite was sent to.                                                                    // 75
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's          // 76
//                       terms of service.                                                                   // 77
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                             // 78
//   storageUsage: Number of bytes this user is currently storing.                                           // 79
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                           // 80
//   appDemoId: If this is an appdemo user (see above), the app ID they started out demoing.                 // 81
//   payments: Object defined by payments module, if loaded.                                                 // 82
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 83
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.             // 84
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                    // 85
                                                                                                             // 86
Packages = new Mongo.Collection("packages");                                                                 // 87
// Packages which are installed or downloading.                                                              // 88
//                                                                                                           // 89
// Each contains:                                                                                            // 90
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 91
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 92
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 93
//       >1 = download byte count.                                                                           // 94
//   error:  If status is "failed", error message string.                                                    // 95
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 96
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 97
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 98
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 99
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 100
//       collector should at some point check whether there are any other references and, if not,            // 101
//       delete the package.                                                                                 // 102
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 103
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 104
//     even if it has no users.                                                                              // 105
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 106
//     packager.                                                                                             // 107
                                                                                                             // 108
DevPackages = new Mongo.Collection("devpackages");                                                           // 109
// List of packages currently made available via the dev tools running on the local machine.                 // 110
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 111
// on the local machine to publish an under-development app to this server. That should only ever            // 112
// happen on developers' desktop machines.                                                                   // 113
//                                                                                                           // 114
// While a dev package is published, it automatically appears as installed by every user of the              // 115
// server, and it overrides all packages with the same application ID. If any instances of those             // 116
// packages are currently open, they are killed and reset on publish.                                        // 117
//                                                                                                           // 118
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 119
// are again killed and refreshed.                                                                           // 120
//                                                                                                           // 121
// Each contains:                                                                                            // 122
//   _id:  The package ID string (as with Packages._id).                                                     // 123
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 124
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 125
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 126
//     changes are made to the source code.                                                                  // 127
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 128
                                                                                                             // 129
UserActions = new Mongo.Collection("userActions");                                                           // 130
// List of actions that each user has installed which create new grains.  Each app may install               // 131
// some number of actions (usually, one).                                                                    // 132
//                                                                                                           // 133
// Each contains:                                                                                            // 134
//   _id:  random                                                                                            // 135
//   userId:  User who has installed this action.                                                            // 136
//   packageId:  Package used to run this action.                                                            // 137
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 138
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 139
//       that clients can access it without subscribing to the Packages collection.                          // 140
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 141
//       searchability.                                                                                      // 142
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 143
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 144
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 145
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 146
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 147
                                                                                                             // 148
Grains = new Mongo.Collection("grains");                                                                     // 149
// Grains belonging to users.                                                                                // 150
//                                                                                                           // 151
// Each contains:                                                                                            // 152
//   _id:  random                                                                                            // 153
//   packageId:  _id of the package of which this grain is an instance.                                      // 154
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 155
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 156
//       searchability.                                                                                      // 157
//   userId: The _id of the account that owns this grain.                                                    // 158
//   identityId: The identity with which the owning account prefers to open this grain.                      // 159
//   title:  Human-readable string title, as chosen by the user.                                             // 160
//   lastUsed:  Date when the grain was last used by a user.                                                 // 161
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 162
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 163
//                   time a session to this grain was opened.                                                // 164
//                                                                                                           // 165
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 166
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 167
// is implemented.                                                                                           // 168
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 169
//       web publishing. This field is initialized when first requested by the app.                          // 170
                                                                                                             // 171
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 172
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 173
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 174
// collection.                                                                                               // 175
                                                                                                             // 176
Contacts = new Mongo.Collection("contacts");                                                                 // 177
// Edges in the social graph.                                                                                // 178
//                                                                                                           // 179
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 180
// will show up in her user-picker UI for actions like share-by-identity.                                    // 181
//                                                                                                           // 182
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 183
// Bob's.                                                                                                    // 184
//                                                                                                           // 185
// Each contains:                                                                                            // 186
//   _id: random                                                                                             // 187
//   ownerId: The `_id` of the user who owns this contact.                                                   // 188
//   identityId:  The identity of the contacted user.                                                        // 189
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 190
//            the contact to the owner.                                                                      // 191
//   created: Date when this contact was created.                                                            // 192
                                                                                                             // 193
Sessions = new Mongo.Collection("sessions");                                                                 // 194
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 195
//                                                                                                           // 196
// Each contains:                                                                                            // 197
//   _id:  random                                                                                            // 198
//   grainId:  _id of the grain to which this session is connected.                                          // 199
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 200
//       '*' in WILDCARD_HOST.                                                                               // 201
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 202
//       period.                                                                                             // 203
//   userId:  User ID of the user who owns this session.                                                     // 204
//   identityId:  Identity ID of the user who owns this session.                                             // 205
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 206
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 207
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 208
//       are present.                                                                                        // 209
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 210
//       interaction.                                                                                        // 211
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 212
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 213
//       until newSession is called on the UiView.                                                           // 214
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 215
//       until newSession is called on the UiView.                                                           // 216
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 217
//       request                                                                                             // 218
                                                                                                             // 219
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 220
// Invite keys which may be used by users to get access to Sandstorm.                                        // 221
//                                                                                                           // 222
// Each contains:                                                                                            // 223
//   _id:  random                                                                                            // 224
//   used:  Boolean indicating whether this key has already been consumed.                                   // 225
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 226
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 227
                                                                                                             // 228
ActivityStats = new Mongo.Collection("activityStats");                                                       // 229
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 230
//                                                                                                           // 231
// Each contains:                                                                                            // 232
//   timestamp: Date when measurements were taken.                                                           // 233
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 234
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 235
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 236
//                                                                                                           // 237
// Each of daily, weekly, and monthly contains:                                                              // 238
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 239
//       interval. Only counts logged-in users.                                                              // 240
//   demoUsers: Demo users.                                                                                  // 241
//   appDemoUsers: Users that came in through "app demo".                                                    // 242
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 243
//   apps: An object indexed by app ID recording, for each app:                                              // 244
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 245
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 246
//         only grains that still exist).                                                                    // 247
//       grains: Number of active grains of this app (that still exist).                                     // 248
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 249
//       demoed: Number of demo grains created and expired.                                                  // 250
//       appDemoUsers: Number of app demos initiated with this app.                                          // 251
                                                                                                             // 252
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 253
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 254
//                                                                                                           // 255
// Each contains:                                                                                            // 256
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 257
//   lastActive: Date of the user's or grain's last activity.                                                // 258
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 259
//     arrived to demo. For others, undefined.                                                               // 260
                                                                                                             // 261
FileTokens = new Mongo.Collection("fileTokens");                                                             // 262
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 263
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 264
// grains) and may use the token to read the file (either to download it, or to restore a new                // 265
// grain from it).                                                                                           // 266
//                                                                                                           // 267
// Each contains:                                                                                            // 268
//   _id:       The unguessable token string.                                                                // 269
//   name:      Suggested filename.                                                                          // 270
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 271
                                                                                                             // 272
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 273
// Access tokens for APIs exported by apps.                                                                  // 274
//                                                                                                           // 275
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 276
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 277
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 278
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and               // 279
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 280
//                                                                                                           // 281
// Each contains:                                                                                            // 282
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 283
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 284
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 285
//              That is, the UiView's newSession() method will intersect the requested permissions           // 286
//              with this identity's permissions before forwarding on to the underlying app. If              // 287
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 288
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 289
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 290
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 291
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 292
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account          // 293
//              that created the token.                                                                      // 294
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 295
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 296
//              permissions will be intersected with those held by `identityId` when the view is             // 297
//              opened.                                                                                      // 298
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 299
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 300
//              on the permissions granted.                                                                  // 301
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 302
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 303
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 304
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 305
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 306
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 307
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 308
//              containing exactly one of the following fields:                                              // 309
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 310
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 311
//                           `Notifications` collection.                                                     // 312
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 313
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 314
//                  database when that happens. This field contains the boolean true to signify that         // 315
//                  it has been set.                                                                         // 316
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 317
//   parentToken: If present, then this token represents exactly the capability represented by               // 318
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 319
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 320
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`             // 321
//              is set to the identity that shared the view, and `accountId` is set to the account           // 322
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                   // 323
//              `parentToken` is present.                                                                    // 324
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 325
//              revocation. This should be displayed when visualizing incoming capabilities to               // 326
//              the grain identified by `grainId`.                                                           // 327
//   created:   Date when this token was created.                                                            // 328
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 329
//              become un-revoked in the future.                                                             // 330
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 331
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)                // 332
//              as passed to the `save()` call that created this token. If not present, treat                // 333
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 334
//   expiresIfUnused:                                                                                        // 335
//              Optional Date after which the token, if it has not been used yet, expires.                   // 336
//              This field should be cleared on a token's first use.                                         // 337
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 338
//              Semantically, this list specifies the powers which were *used* to originally                 // 339
//              create the token. If any condition in the list becomes untrue, then the token must           // 340
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 341
//              through it must also become revoked. Each item is the JSON serialization of the              // 342
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 343
//                                                                                                           // 344
// It is important to note that a token's owner and provider are independent from each other. To             // 345
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 346
//                                                                                                           // 347
// struct ApiToken {                                                                                         // 348
//   owner :ApiTokenOwner;                                                                                   // 349
//   provider :union {                                                                                       // 350
//     grain :group {                                                                                        // 351
//       grainId :Text;                                                                                      // 352
//       union {                                                                                             // 353
//         uiView :group {                                                                                   // 354
//           identityId :Text;                                                                               // 355
//           roleAssignment :RoleAssignment;                                                                 // 356
//           forSharing :Bool;                                                                               // 357
//         }                                                                                                 // 358
//         objectId :SupervisorObjectId;                                                                     // 359
//       }                                                                                                   // 360
//     }                                                                                                     // 361
//     frontendRef :union {                                                                                  // 362
//        notificationHandle :Text;                                                                          // 363
//        ipNetwork :Bool;                                                                                   // 364
//        ipInterface :Bool;                                                                                 // 365
//     }                                                                                                     // 366
//     child :group {                                                                                        // 367
//       parentToken :Text;                                                                                  // 368
//       union {                                                                                             // 369
//         uiView :group {                                                                                   // 370
//           grainId :Text;                                                                                  // 371
//           identityId :Text;                                                                               // 372
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 373
//         }                                                                                                 // 374
//         other :Void;                                                                                      // 375
//       }                                                                                                   // 376
//     }                                                                                                     // 377
//   }                                                                                                       // 378
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 379
//   ...                                                                                                     // 380
// }                                                                                                         // 381
                                                                                                             // 382
Notifications = new Mongo.Collection("notifications");                                                       // 383
// Notifications for a user.                                                                                 // 384
//                                                                                                           // 385
// Each contains:                                                                                            // 386
//   _id:          random                                                                                    // 387
//   grainId:      The grain originating this notification, if any.                                          // 388
//   userId:       The user receiving the notification.                                                      // 389
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 390
//   isUnread:     Boolean indicating if this notification is unread.                                        // 391
//   timestamp:    Date when this notification was last updated                                              // 392
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 393
//                 ApiToken referencing the `OngoingNotification` capability.                                // 394
//   admin:        If present, this is a notification intended for an admin.                                 // 395
//     action:     If present, this is a (string) link that the notification should direct the               // 396
//                 admin to.                                                                                 // 397
//     type:       The type of notification (currently only "reportStats").                                  // 398
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 399
//                 as keys.                                                                                  // 400
//     $appId:     The appId that has an outstanding update.                                                 // 401
//       packageId: The packageId that it will update to.                                                    // 402
//       name: The name of the app. (appTitle from package.manifest)                                         // 403
//       version: The app's version number. (appVersion from package.manifest)                               // 404
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)
                                                                                                             // 406
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 407
// Access tokens for the Stats collection                                                                    // 408
//                                                                                                           // 409
// These tokens are used for accessing the ActivityStats collection remotely                                 // 410
// (ie. from a dashboard webapp)                                                                             // 411
//                                                                                                           // 412
// Each contains:                                                                                            // 413
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 414
                                                                                                             // 415
Misc = new Mongo.Collection("misc");                                                                         // 416
// Miscellaneous configuration and other settings                                                            // 417
//                                                                                                           // 418
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 419
// but in general any miscellaneous settings should go in here                                               // 420
//                                                                                                           // 421
// Each contains:                                                                                            // 422
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 423
//   value:     The value of the setting.                                                                    // 424
                                                                                                             // 425
Settings = new Mongo.Collection("settings");                                                                 // 426
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 427
// route. This collection differs from misc in that this collection is completely user controlled.           // 428
//                                                                                                           // 429
// Each contains:                                                                                            // 430
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 431
//   value:     The value of the setting.                                                                    // 432
//   potentially other fields that are unique to the setting                                                 // 433
                                                                                                             // 434
Migrations = new Mongo.Collection("migrations");                                                             // 435
// This table tracks which migrations we have applied to this instance.                                      // 436
// It contains a single entry:                                                                               // 437
//   _id:       "migrations_applied"                                                                         // 438
//   value:     The number of migrations this instance has successfully completed.                           // 439
                                                                                                             // 440
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 441
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 442
// support relatively small assets: under 1MB each.                                                          // 443
//                                                                                                           // 444
// Each contains:                                                                                            // 445
//   _id:       Random ID; will be used in the URL.                                                          // 446
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 447
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 448
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 449
//   content:   The asset content (byte buffer).                                                             // 450
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 451
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 452
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 453
                                                                                                             // 454
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 455
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 456
// profile picture.                                                                                          // 457
//                                                                                                           // 458
// Each contains:                                                                                            // 459
//   _id:       Random ID.                                                                                   // 460
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 461
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 462
//           userId: User whose picture shall be replaced.                                                   // 463
//           identityId: Which of the user's identities shall be updated.                                    // 464
//   expires:   Time when this token will go away if unused.                                                 // 465
                                                                                                             // 466
Plans = new Mongo.Collection("plans");                                                                       // 467
// Subscription plans, which determine quota.                                                                // 468
//                                                                                                           // 469
// Each contains:                                                                                            // 470
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 471
//   storage: Number of bytes this user is allowed to store.                                                 // 472
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 473
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 474
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 475
//   price: Price per month in US cents.                                                                     // 476
                                                                                                             // 477
AppIndex = new Mongo.Collection("appIndex");                                                                 // 478
// A mirror of the data from the App Market index                                                            // 479
//                                                                                                           // 480
// Each contains:                                                                                            // 481
//   _id: the appId of the app                                                                               // 482
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 483
                                                                                                             // 484
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 485
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 486
// by that user is installed, as well as if the keybase profile is requested and not already                 // 487
// present for some reason.                                                                                  // 488
//                                                                                                           // 489
// Each contains:                                                                                            // 490
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 491
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 492
//   handle: Keybase handle.                                                                                 // 493
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 494
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 495
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 496
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 497
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 498
//     server is currently actively checking this proof). Note that if a check fails due to network          // 499
//     errors, the status goes back to "unverified".                                                         // 500
//                                                                                                           // 501
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 502
//       for now and we just trust Keybase.                                                                  // 503
                                                                                                             // 504
if (Meteor.isServer) {                                                                                       // 505
  Meteor.publish("credentials", function () {                                                                // 506
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 507
                                                                                                             // 508
    if (this.userId) {                                                                                       // 509
      return [                                                                                               // 510
        Meteor.users.find({_id: this.userId},                                                                // 511
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 512
                      plan: 1, hasCompletedSignup: 1}}),                                                     // 513
        Plans.find()                                                                                         // 514
      ];                                                                                                     // 515
    } else {                                                                                                 // 516
      return [];                                                                                             // 517
    }                                                                                                        // 518
  });                                                                                                        // 519
}                                                                                                            // 520
                                                                                                             // 521
isDemoUser = function() {                                                                                    // 522
  // Returns true if this is a demo user.                                                                    // 523
                                                                                                             // 524
  var user = Meteor.user();                                                                                  // 525
  if (user && user.expires) {                                                                                // 526
    return true;                                                                                             // 527
  } else {                                                                                                   // 528
    return false;                                                                                            // 529
  }                                                                                                          // 530
}                                                                                                            // 531
                                                                                                             // 532
isSignedUp = function() {                                                                                    // 533
  // Returns true if the user has presented an invite key.                                                   // 534
                                                                                                             // 535
  var user = Meteor.user();                                                                                  // 536
                                                                                                             // 537
  if (!user) return false;  // not signed in                                                                 // 538
                                                                                                             // 539
  if (user.expires) return false;  // demo user.                                                             // 540
                                                                                                             // 541
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 542
                                                                                                             // 543
  if (user.signupKey) return true;  // user is invited                                                       // 544
                                                                                                             // 545
  return false;                                                                                              // 546
}                                                                                                            // 547
                                                                                                             // 548
isSignedUpOrDemo = function () {                                                                             // 549
  var user = Meteor.user();                                                                                  // 550
                                                                                                             // 551
  if (!user) return false;  // not signed in                                                                 // 552
                                                                                                             // 553
  if (user.expires) return true;  // demo user.                                                              // 554
                                                                                                             // 555
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 556
                                                                                                             // 557
  if (user.signupKey) return true;  // user is invited                                                       // 558
                                                                                                             // 559
  return false;                                                                                              // 560
}                                                                                                            // 561
                                                                                                             // 562
isUserOverQuota = function (user) {                                                                          // 563
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 564
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 565
  // which may still increase their storage usage.                                                           // 566
  //                                                                                                         // 567
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 568
                                                                                                             // 569
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 570
                                                                                                             // 571
  var plan = Plans.findOne(user.plan || "free");                                                             // 572
                                                                                                             // 573
  if (plan.grains < Infinity) {                                                                              // 574
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 575
    if (count >= plan.grains) return "outOfGrains";                                                          // 576
  }                                                                                                          // 577
                                                                                                             // 578
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 579
}                                                                                                            // 580
                                                                                                             // 581
isUserExcessivelyOverQuota = function (user) {                                                               // 582
  // Return true if user is so far over quota that we should prevent their existing grains from              // 583
  // running at all.                                                                                         // 584
  //                                                                                                         // 585
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 586
                                                                                                             // 587
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 588
                                                                                                             // 589
  var plan = Plans.findOne(user.plan || "free");                                                             // 590
                                                                                                             // 591
  if (plan.grains < Infinity) {                                                                              // 592
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains * 2}).count();               // 593
    if (count >= plan.grains * 2) return "outOfGrains";                                                      // 594
  }                                                                                                          // 595
                                                                                                             // 596
  return plan && user.storageUsage && user.storageUsage >= plan.storage * 1.2 && "outOfStorage";             // 597
}                                                                                                            // 598
                                                                                                             // 599
isAdmin = function() {                                                                                       // 600
  // Returns true if the user is the administrator.                                                          // 601
                                                                                                             // 602
  var user = Meteor.user();                                                                                  // 603
  if (user && user.isAdmin) {                                                                                // 604
    return true;                                                                                             // 605
  } else {                                                                                                   // 606
    return false;                                                                                            // 607
  }                                                                                                          // 608
}                                                                                                            // 609
                                                                                                             // 610
isAdminById = function(id) {                                                                                 // 611
  // Returns true if the user's id is the administrator.                                                     // 612
                                                                                                             // 613
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 614
  if (user && user.isAdmin) {                                                                                // 615
    return true;                                                                                             // 616
  } else {                                                                                                   // 617
    return false;                                                                                            // 618
  }                                                                                                          // 619
}                                                                                                            // 620
                                                                                                             // 621
findAdminUserForToken = function (token) {                                                                   // 622
  if (!token.requirements) {                                                                                 // 623
    return;                                                                                                  // 624
  }                                                                                                          // 625
  var requirements = token.requirements.filter(function (requirement) {                                      // 626
    return "userIsAdmin" in requirement;                                                                     // 627
  });                                                                                                        // 628
                                                                                                             // 629
  if (requirements.length > 1) {                                                                             // 630
    return;                                                                                                  // 631
  }                                                                                                          // 632
  if (requirements.length === 0) {                                                                           // 633
    return;                                                                                                  // 634
  }                                                                                                          // 635
  return requirements[0].userIsAdmin;                                                                        // 636
};                                                                                                           // 637
                                                                                                             // 638
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 639
                                                                                                             // 640
if (wildcardHost.length != 2) {                                                                              // 641
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 642
}                                                                                                            // 643
                                                                                                             // 644
matchWildcardHost = function(host) {                                                                         // 645
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 646
                                                                                                             // 647
  var prefix = wildcardHost[0];                                                                              // 648
  var suffix = wildcardHost[1];                                                                              // 649
                                                                                                             // 650
  // We remove everything after the first ":" character so that our                                          // 651
  // comparison logic ignores port numbers.                                                                  // 652
  suffix = suffix.split(":")[0];                                                                             // 653
  host = host.split(":")[0];                                                                                 // 654
                                                                                                             // 655
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 656
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 657
      host.length >= prefix.length + suffix.length) {                                                        // 658
    var id = host.slice(prefix.length, -suffix.length);                                                      // 659
    if (id.match(/^[a-z0-9]*$/)) {                                                                           // 660
      return id;                                                                                             // 661
    }                                                                                                        // 662
  }                                                                                                          // 663
                                                                                                             // 664
  return null;                                                                                               // 665
}                                                                                                            // 666
                                                                                                             // 667
makeWildcardHost = function (id) {                                                                           // 668
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 669
}                                                                                                            // 670
                                                                                                             // 671
if (Meteor.isServer) {                                                                                       // 672
  var Url = Npm.require("url");                                                                              // 673
  getWildcardOrigin = function () {                                                                          // 674
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 675
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 676
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 677
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 678
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 679
                                                                                                             // 680
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 681
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 682
                                                                                                             // 683
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 684
    if (dotPos < 0) {                                                                                        // 685
      return protocol + "//*";                                                                               // 686
    } else {                                                                                                 // 687
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 688
    }                                                                                                        // 689
  }                                                                                                          // 690
}                                                                                                            // 691
                                                                                                             // 692
allowDevAccounts = function () {                                                                             // 693
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 694
  if (setting) {                                                                                             // 695
    return setting.value;                                                                                    // 696
  } else {                                                                                                   // 697
    return Meteor.settings && Meteor.settings.public &&                                                      // 698
           Meteor.settings.public.allowDevAccounts;                                                          // 699
  }                                                                                                          // 700
};                                                                                                           // 701
                                                                                                             // 702
roleAssignmentPattern = {                                                                                    // 703
  none : Match.Optional(null),                                                                               // 704
  allAccess: Match.Optional(null),                                                                           // 705
  roleId: Match.Optional(Match.Integer),                                                                     // 706
  addPermissions: Match.Optional([Boolean]),                                                                 // 707
  removePermissions: Match.Optional([Boolean]),                                                              // 708
};                                                                                                           // 709
                                                                                                             // 710
SandstormDb = function () {                                                                                  // 711
  this.collections = {                                                                                       // 712
    // Direct access to underlying collections. DEPRECATED.                                                  // 713
    //                                                                                                       // 714
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 715
    //   direct access to the collections.                                                                   // 716
                                                                                                             // 717
    packages: Packages,                                                                                      // 718
    devPackages: DevPackages,                                                                                // 719
    userActions: UserActions,                                                                                // 720
    grains: Grains,                                                                                          // 721
    contacts: Contacts,                                                                                      // 722
    sessions: Sessions,                                                                                      // 723
    signupKeys: SignupKeys,                                                                                  // 724
    activityStats: ActivityStats,                                                                            // 725
    deleteStats: DeleteStats,                                                                                // 726
    fileTokens: FileTokens,                                                                                  // 727
    apiTokens: ApiTokens,                                                                                    // 728
    notifications: Notifications,                                                                            // 729
    statsTokens: StatsTokens,                                                                                // 730
    misc: Misc,                                                                                              // 731
    settings: Settings,                                                                                      // 732
    appIndex: AppIndex,                                                                                      // 733
    keybaseProfiles: KeybaseProfiles,                                                                        // 734
                                                                                                             // 735
    // Intentionally omitted:                                                                                // 736
    // - Migrations, since it's used only within this package.                                               // 737
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 738
  };                                                                                                         // 739
};                                                                                                           // 740
                                                                                                             // 741
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 742
//   objects created in SandstormDb's constructor rather than globals.                                       // 743
                                                                                                             // 744
_.extend(SandstormDb.prototype, {                                                                            // 745
  isDemoUser: isDemoUser,                                                                                    // 746
  isSignedUp: isSignedUp,                                                                                    // 747
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 748
  isUserOverQuota: isUserOverQuota,                                                                          // 749
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 750
  isAdmin: isAdmin,                                                                                          // 751
  isAdminById: isAdminById,                                                                                  // 752
  findAdminUserForToken: findAdminUserForToken,                                                              // 753
  matchWildcardHost: matchWildcardHost,                                                                      // 754
  makeWildcardHost: makeWildcardHost,                                                                        // 755
  allowDevAccounts: allowDevAccounts,                                                                        // 756
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 757
});                                                                                                          // 758
                                                                                                             // 759
if (Meteor.isServer) {                                                                                       // 760
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 761
}                                                                                                            // 762
                                                                                                             // 763
// =======================================================================================                   // 764
// Below this point are newly-written or refactored functions.                                               // 765
                                                                                                             // 766
_.extend(SandstormDb.prototype, {                                                                            // 767
  getUser: function getUser (userId) {                                                                       // 768
    check(userId, Match.OneOf(String, undefined, null));                                                     // 769
    if (userId) {                                                                                            // 770
      return Meteor.users.findOne(userId);                                                                   // 771
    }                                                                                                        // 772
  },                                                                                                         // 773
                                                                                                             // 774
  getIdentity: function getIdentity (identityId) {                                                           // 775
    check(identityId, String);                                                                               // 776
    var identity = Meteor.users.findOne({_id: identityId});                                                  // 777
    if (identity) {                                                                                          // 778
      SandstormDb.fillInProfileDefaults(identity);                                                           // 779
      SandstormDb.fillInIntrinsicName(identity);                                                             // 780
      SandstormDb.fillInPictureUrl(identity);                                                                // 781
      return identity;                                                                                       // 782
    }                                                                                                        // 783
  },                                                                                                         // 784
                                                                                                             // 785
  userHasIdentity: function (userId, identityId) {                                                           // 786
    check(userId, String);                                                                                   // 787
    check(identityId, String);                                                                               // 788
                                                                                                             // 789
    if (userId === identityId) return true;                                                                  // 790
                                                                                                             // 791
    var user = Meteor.users.findOne(userId);                                                                 // 792
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                   // 793
  },                                                                                                         // 794
                                                                                                             // 795
  userGrains: function userGrains (userId) {                                                                 // 796
    check(userId, Match.OneOf(String, undefined, null));                                                     // 797
    return this.collections.grains.find({userId: userId});                                                   // 798
  },                                                                                                         // 799
                                                                                                             // 800
  currentUserGrains: function currentUserGrains () {                                                         // 801
    return this.userGrains(Meteor.userId());                                                                 // 802
  },                                                                                                         // 803
                                                                                                             // 804
  getGrain: function getGrain (grainId) {                                                                    // 805
    check(grainId, String);                                                                                  // 806
    return this.collections.grains.findOne(grainId);                                                         // 807
  },                                                                                                         // 808
                                                                                                             // 809
  userApiTokens: function userApiTokens (userId) {                                                           // 810
    check(userId, Match.OneOf(String, undefined, null));                                                     // 811
    var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                  // 812
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 813
  },                                                                                                         // 814
                                                                                                             // 815
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 816
    return this.userApiTokens(Meteor.userId());                                                              // 817
  },                                                                                                         // 818
                                                                                                             // 819
  userActions: function userActions (user) {                                                                 // 820
    return this.collections.userActions.find({userId: user});                                                // 821
  },                                                                                                         // 822
                                                                                                             // 823
  currentUserActions: function currentUserActions () {                                                       // 824
    return this.userActions(Meteor.userId());                                                                // 825
  },                                                                                                         // 826
                                                                                                             // 827
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 828
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 829
  },                                                                                                         // 830
                                                                                                             // 831
  getPlan: function (id) {                                                                                   // 832
    check(id, String);                                                                                       // 833
    var plan = Plans.findOne(id);                                                                            // 834
    if (!plan) {                                                                                             // 835
      throw new Error("no such plan: " + id);                                                                // 836
    }                                                                                                        // 837
    return plan;                                                                                             // 838
  },                                                                                                         // 839
                                                                                                             // 840
  listPlans: function () {                                                                                   // 841
    return Plans.find({}, {sort: {price: 1}});                                                               // 842
  },                                                                                                         // 843
                                                                                                             // 844
  getMyPlan: function () {                                                                                   // 845
    var user = Meteor.user();                                                                                // 846
    return user && Plans.findOne(user.plan || "free");                                                       // 847
  },                                                                                                         // 848
                                                                                                             // 849
  getMyUsage: function (user) {                                                                              // 850
    user = user || Meteor.user();                                                                            // 851
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 852
      if (Meteor.isClient) {                                                                                 // 853
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 854
        // not reactive.                                                                                     // 855
        return user.pseudoUsage;                                                                             // 856
      } else {                                                                                               // 857
        return {                                                                                             // 858
          grains: Grains.find({userId: user._id}).count(),                                                   // 859
          storage: user.storageUsage || 0,                                                                   // 860
          compute: 0   // not tracked yet                                                                    // 861
        };                                                                                                   // 862
      }                                                                                                      // 863
    } else {                                                                                                 // 864
      return {grains: 0, storage: 0, compute: 0};                                                            // 865
    }                                                                                                        // 866
  },                                                                                                         // 867
                                                                                                             // 868
  isUninvitedFreeUser: function () {                                                                         // 869
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 870
                                                                                                             // 871
    var user = Meteor.user();                                                                                // 872
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 873
  },                                                                                                         // 874
                                                                                                             // 875
  getSetting: function (name) {                                                                              // 876
    var setting = Settings.findOne(name);                                                                    // 877
    return setting && setting.value;                                                                         // 878
  },                                                                                                         // 879
                                                                                                             // 880
  addUserActions: function(packageId) {                                                                      // 881
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 882
    var pack = Packages.findOne(packageId);                                                                  // 883
    if (pack) {                                                                                              // 884
      // Remove old versions.                                                                                // 885
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 886
          .forEach(function (action) {                                                                       // 887
        UserActions.remove(action._id);                                                                      // 888
      });                                                                                                    // 889
                                                                                                             // 890
      // Install new.                                                                                        // 891
      var actions = pack.manifest.actions;                                                                   // 892
      for (i in actions) {                                                                                   // 893
        var action = actions[i];                                                                             // 894
        if ("none" in action.input) {                                                                        // 895
          var userAction = {                                                                                 // 896
            userId: Meteor.userId(),                                                                         // 897
            packageId: pack._id,                                                                             // 898
            appId: pack.appId,                                                                               // 899
            appTitle: pack.manifest.appTitle,                                                                // 900
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 901
            appVersion: pack.manifest.appVersion,                                                            // 902
            title: action.title,                                                                             // 903
            nounPhrase: action.nounPhrase,                                                                   // 904
            command: action.command                                                                          // 905
          };                                                                                                 // 906
          UserActions.insert(userAction);                                                                    // 907
        } else {                                                                                             // 908
          // TODO(someday):  Implement actions with capability inputs.                                       // 909
        }                                                                                                    // 910
      }                                                                                                      // 911
                                                                                                             // 912
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 913
    }                                                                                                        // 914
  },                                                                                                         // 915
                                                                                                             // 916
  sendAdminNotification: function (message, link) {                                                          // 917
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 918
      Notifications.insert({                                                                                 // 919
        admin: {                                                                                             // 920
          action: link,                                                                                      // 921
          type: "reportStats",                                                                               // 922
        },                                                                                                   // 923
        userId: user._id,                                                                                    // 924
        text: {defaultText: message},                                                                        // 925
        timestamp: new Date(),                                                                               // 926
        isUnread: true,                                                                                      // 927
      });                                                                                                    // 928
    });                                                                                                      // 929
  },                                                                                                         // 930
                                                                                                             // 931
  getKeybaseProfile: function (keyFingerprint) {                                                             // 932
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 933
  },                                                                                                         // 934
});                                                                                                          // 935
                                                                                                             // 936
var appNameFromPackage = function(packageObj) {                                                              // 937
  // This function takes a Package object from Mongo and returns an                                          // 938
  // app title.                                                                                              // 939
  var manifest = packageObj.manifest;                                                                        // 940
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 941
  var action = manifest.actions[0];                                                                          // 942
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 943
    appNameFromActionName(action.title.defaultText);                                                         // 944
  return appName;                                                                                            // 945
};                                                                                                           // 946
                                                                                                             // 947
var appNameFromActionName = function(name) {                                                                 // 948
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 949
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 950
  //   derive the app title from the action title.                                                           // 951
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 952
  if (!name) {                                                                                               // 953
    return "(unnamed)";                                                                                      // 954
  }                                                                                                          // 955
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 956
    name = name.slice(4);                                                                                    // 957
  }                                                                                                          // 958
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 959
    name = "Hacker CMS";                                                                                     // 960
  } else {                                                                                                   // 961
    var space = name.indexOf(" ");                                                                           // 962
    if (space > 0) {                                                                                         // 963
      name = name.slice(0, space);                                                                           // 964
    }                                                                                                        // 965
  }                                                                                                          // 966
  return name;                                                                                               // 967
};                                                                                                           // 968
                                                                                                             // 969
var appShortDescriptionFromPackage = function (pkg) {                                                        // 970
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 971
         pkg.manifest.metadata.shortDescription &&                                                           // 972
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 973
};                                                                                                           // 974
                                                                                                             // 975
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 976
  // A hack to deal with legacy apps not including fields in their manifests.                                // 977
  // I look forward to the day I can remove most of this code.                                               // 978
  // Attempt to figure out the appropriate noun that this action will create.                                // 979
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 980
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 981
  // Otherwise, try to guess one from the structure of the action title field                                // 982
  if (action.title && action.title.defaultText) {                                                            // 983
    var text = action.title.defaultText;                                                                     // 984
    // Strip a leading "New "                                                                                // 985
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 986
      var candidate = text.slice(4);                                                                         // 987
      // Strip a leading appname too, if provided                                                            // 988
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 989
        var newCandidate = candidate.slice(appTitle.length);                                                 // 990
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 991
        if (newCandidate.length > 0) {                                                                       // 992
          return newCandidate.toLowerCase();                                                                 // 993
        } else {                                                                                             // 994
          return "grain";                                                                                    // 995
        }                                                                                                    // 996
      }                                                                                                      // 997
      return candidate.toLowerCase();                                                                        // 998
    }                                                                                                        // 999
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 1000
    // the package soon.                                                                                     // 1001
    return text;                                                                                             // 1002
  } else {                                                                                                   // 1003
    return "grain";                                                                                          // 1004
  }                                                                                                          // 1005
};                                                                                                           // 1006
                                                                                                             // 1007
// Static methods on SandstormDb that don't need an instance.                                                // 1008
// Largely things that deal with backwards-compatibility.                                                    // 1009
_.extend(SandstormDb, {                                                                                      // 1010
  appNameFromActionName: appNameFromActionName,                                                              // 1011
  appNameFromPackage: appNameFromPackage,                                                                    // 1012
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 1013
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 1014
});                                                                                                          // 1015
                                                                                                             // 1016
if (Meteor.isServer) {                                                                                       // 1017
  var Crypto = Npm.require("crypto");                                                                        // 1018
  var ContentType = Npm.require("content-type");                                                             // 1019
  var Zlib = Npm.require("zlib");                                                                            // 1020
                                                                                                             // 1021
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1022
                                                                                                             // 1023
  var computeStagger = function (n) {                                                                        // 1024
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1025
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1026
    // The sequence looks like:                                                                              // 1027
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1028
    //                                                                                                       // 1029
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1030
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1031
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1032
    // earlier ones schedule their events.                                                                   // 1033
    var denom = 1;                                                                                           // 1034
    while (denom <= n) denom <<= 1;                                                                          // 1035
    var num = n * 2 - denom + 1;                                                                             // 1036
    return num / denom;                                                                                      // 1037
  }                                                                                                          // 1038
                                                                                                             // 1039
  var stagger = computeStagger(replicaNumber);                                                               // 1040
                                                                                                             // 1041
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1042
    // Register a database cleanup function than should run periodically, roughly once every                 // 1043
    // interval of the given length.                                                                         // 1044
    //                                                                                                       // 1045
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1046
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1047
    // frequent cleanups.                                                                                    // 1048
                                                                                                             // 1049
    check(intervalMs, Number);                                                                               // 1050
    check(callback, Function);                                                                               // 1051
                                                                                                             // 1052
    if (intervalMs < 120000) {                                                                               // 1053
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1054
                      "are you using the right units?");                                                     // 1055
    }                                                                                                        // 1056
                                                                                                             // 1057
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1058
    // the schedule is independent of the exact startup time.                                                // 1059
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1060
                                                                                                             // 1061
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1062
    // same data at the same time.                                                                           // 1063
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1064
                                                                                                             // 1065
    // If the stagger put us more than an interval away from now, back up.                                   // 1066
    if (first > intervalMs) first -= intervalMs;                                                             // 1067
                                                                                                             // 1068
    Meteor.setTimeout(function () {                                                                          // 1069
      callback();                                                                                            // 1070
      Meteor.setInterval(callback, intervalMs);                                                              // 1071
    }, first);                                                                                               // 1072
  }                                                                                                          // 1073
                                                                                                             // 1074
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1075
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1076
                                                                                                             // 1077
  var BufferSmallerThan = function (limit) {                                                                 // 1078
    return Match.Where(function (buf) {                                                                      // 1079
      check(buf, Buffer);                                                                                    // 1080
      return buf.length < limit;                                                                             // 1081
    });                                                                                                      // 1082
  }                                                                                                          // 1083
                                                                                                             // 1084
  var DatabaseId = Match.Where(function (s) {                                                                // 1085
    check(s, String);                                                                                        // 1086
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1087
  });                                                                                                        // 1088
                                                                                                             // 1089
  addStaticAsset = function (metadata, content) {                                                            // 1090
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1091
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1092
                                                                                                             // 1093
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1094
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1095
      metadata.encoding = "gzip";                                                                            // 1096
    }                                                                                                        // 1097
                                                                                                             // 1098
    check(metadata, {                                                                                        // 1099
      mimeType: String,                                                                                      // 1100
      encoding: Match.Optional("gzip")                                                                       // 1101
    });                                                                                                      // 1102
    check(content, BufferSmallerThan(1 << 20));                                                              // 1103
                                                                                                             // 1104
    // Validate content type.                                                                                // 1105
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1106
                                                                                                             // 1107
    var hasher = Crypto.createHash("sha256");                                                                // 1108
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1109
    hasher.update(content);                                                                                  // 1110
    var hash = hasher.digest("base64");                                                                      // 1111
                                                                                                             // 1112
    var existing = StaticAssets.findAndModify({                                                              // 1113
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1114
      update: {$inc: {refcount: 1}},                                                                         // 1115
      fields: {_id: 1, refcount: 1},                                                                         // 1116
    });                                                                                                      // 1117
    if (existing) {                                                                                          // 1118
      return existing._id;                                                                                   // 1119
    }                                                                                                        // 1120
                                                                                                             // 1121
    return StaticAssets.insert(_.extend({                                                                    // 1122
      hash: hash,                                                                                            // 1123
      content: content,                                                                                      // 1124
      refcount: 1                                                                                            // 1125
    }, metadata));                                                                                           // 1126
  }                                                                                                          // 1127
                                                                                                             // 1128
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1129
                                                                                                             // 1130
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1131
    // Increment the refcount on an existing static asset.                                                   // 1132
    //                                                                                                       // 1133
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1134
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1135
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1136
                                                                                                             // 1137
    check(id, String);                                                                                       // 1138
                                                                                                             // 1139
    var existing = StaticAssets.findAndModify({                                                              // 1140
      query: {hash: hash},                                                                                   // 1141
      update: {$inc: {refcount: 1}},                                                                         // 1142
      fields: {_id: 1, refcount: 1},                                                                         // 1143
    });                                                                                                      // 1144
    if (!existing) {                                                                                         // 1145
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1146
    }                                                                                                        // 1147
  }                                                                                                          // 1148
                                                                                                             // 1149
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1150
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1151
    //                                                                                                       // 1152
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1153
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1154
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1155
                                                                                                             // 1156
    check(id, String);                                                                                       // 1157
                                                                                                             // 1158
    var existing = StaticAssets.findAndModify({                                                              // 1159
      query: {_id: id},                                                                                      // 1160
      update: {$inc: {refcount: -1}},                                                                        // 1161
      fields: {_id: 1, refcount: 1},                                                                         // 1162
      new: true,                                                                                             // 1163
    });                                                                                                      // 1164
    if (!existing) {                                                                                         // 1165
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1166
    } else if (existing.refcount <= 0) {                                                                     // 1167
      StaticAssets.remove({_id: existing._id});                                                              // 1168
    }                                                                                                        // 1169
  }                                                                                                          // 1170
                                                                                                             // 1171
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1172
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1173
                                                                                                             // 1174
    check(id, String);                                                                                       // 1175
                                                                                                             // 1176
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1177
    if (asset) {                                                                                             // 1178
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1179
      //   here.                                                                                             // 1180
      asset.content = new Buffer(asset.content);                                                             // 1181
    }                                                                                                        // 1182
    return asset;                                                                                            // 1183
  }                                                                                                          // 1184
                                                                                                             // 1185
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1186
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1187
                                                                                                             // 1188
    return AssetUploadTokens.insert({                                                                        // 1189
      purpose: purpose,                                                                                      // 1190
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1191
    });                                                                                                      // 1192
  }                                                                                                          // 1193
                                                                                                             // 1194
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1195
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1196
    // returned. If no matching upload exists, returns undefined.                                            // 1197
                                                                                                             // 1198
    check(id, String);                                                                                       // 1199
                                                                                                             // 1200
    var upload = AssetUploadTokens.findAndModify({                                                           // 1201
      query: {_id: id},                                                                                      // 1202
      remove: true                                                                                           // 1203
    });                                                                                                      // 1204
                                                                                                             // 1205
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1206
      return undefined;  // already expired                                                                  // 1207
    } else {                                                                                                 // 1208
      return upload.purpose;                                                                                 // 1209
    }                                                                                                        // 1210
  }                                                                                                          // 1211
                                                                                                             // 1212
  function cleanupExpiredAssetUploads() {                                                                    // 1213
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1214
  }                                                                                                          // 1215
                                                                                                             // 1216
  // Cleanup tokens every hour.                                                                              // 1217
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1218
                                                                                                             // 1219
  var packageCache = {};                                                                                     // 1220
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1221
                                                                                                             // 1222
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1223
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1224
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1225
    // lot.                                                                                                  // 1226
                                                                                                             // 1227
    if (packageId in packageCache) {                                                                         // 1228
      return packageCache[packageId];                                                                        // 1229
    }                                                                                                        // 1230
                                                                                                             // 1231
    var pkg = Packages.findOne(packageId);                                                                   // 1232
    if (pkg && pkg.status === "ready") {                                                                     // 1233
      packageCache[packageId] = pkg;                                                                         // 1234
    }                                                                                                        // 1235
    return pkg;                                                                                              // 1236
  }                                                                                                          // 1237
                                                                                                             // 1238
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1239
                                                               versionNumber, marketingVersion) {            // 1240
    var db = this;                                                                                           // 1241
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1242
      {fields: {userId: 1}});                                                                                // 1243
    actions.forEach(function (action) {                                                                      // 1244
      var userId = action.userId;                                                                            // 1245
      var updater = {                                                                                        // 1246
        userId: userId,                                                                                      // 1247
        timestamp: new Date(),                                                                               // 1248
        isUnread: true,                                                                                      // 1249
      };                                                                                                     // 1250
                                                                                                             // 1251
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1252
      // field inside of an object to update                                                                 // 1253
      updater["appUpdates." + appId] = {                                                                     // 1254
        marketingVersion: marketingVersion,                                                                  // 1255
        packageId: packageId,                                                                                // 1256
        name: name,                                                                                          // 1257
        version: versionNumber,                                                                              // 1258
      };                                                                                                     // 1259
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1260
    });                                                                                                      // 1261
                                                                                                             // 1262
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1263
                                                                                                             // 1264
    // In the case where we replaced a previous notification and that was the only reference to the          // 1265
    // package, we need to clean it up                                                                       // 1266
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1267
  };                                                                                                         // 1268
                                                                                                             // 1269
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1270
    check(appId, String);                                                                                    // 1271
    check(version, Match.Integer);                                                                           // 1272
    check(packageId, String);                                                                                // 1273
                                                                                                             // 1274
    var selector = {                                                                                         // 1275
      userId: Meteor.userId(),                                                                               // 1276
      appId: appId,                                                                                          // 1277
      appVersion: { $lte: version },                                                                         // 1278
      packageId: { $ne: packageId }                                                                          // 1279
    };                                                                                                       // 1280
                                                                                                             // 1281
    if (!this.isSimulation) {                                                                                // 1282
      Grains.find(selector).forEach(function (grain) {                                                       // 1283
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1284
      });                                                                                                    // 1285
    }                                                                                                        // 1286
                                                                                                             // 1287
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1288
  };                                                                                                         // 1289
                                                                                                             // 1290
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1291
    // Mark package for possible installation.                                                               // 1292
                                                                                                             // 1293
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1294
                                                                                                             // 1295
    if (retryFailed) {                                                                                       // 1296
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1297
    } else {                                                                                                 // 1298
      try {                                                                                                  // 1299
        fields._id = packageId;                                                                              // 1300
        Packages.insert(fields);                                                                             // 1301
      } catch (err) {                                                                                        // 1302
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1303
      }                                                                                                      // 1304
    }                                                                                                        // 1305
  };                                                                                                         // 1306
                                                                                                             // 1307
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1308
    check(keyFingerprint, String);                                                                           // 1309
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1310
  });                                                                                                        // 1311
                                                                                                             // 1312
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1313
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1314
                                                                                                             // 1315
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1316
                                                                                                             // 1317
    console.log("fetching keybase", keyFingerprint);                                                         // 1318
                                                                                                             // 1319
    HTTP.get(                                                                                                // 1320
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1321
        "&fields=basics,profile,proofs_summary", {                                                           // 1322
      timeout: 5000                                                                                          // 1323
    }, function (err, keybaseResponse) {                                                                     // 1324
      if (err) {                                                                                             // 1325
        console.log("keybase lookup error:", err.stack);                                                     // 1326
        return;                                                                                              // 1327
      }                                                                                                      // 1328
                                                                                                             // 1329
      if (!keybaseResponse.data) {                                                                           // 1330
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1331
        return;                                                                                              // 1332
      }                                                                                                      // 1333
                                                                                                             // 1334
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1335
                                                                                                             // 1336
      if (profile) {                                                                                         // 1337
        var record = {                                                                                       // 1338
          displayName: (profile.profile || {}).full_name,                                                    // 1339
          handle: (profile.basics || {}).username,                                                           // 1340
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1341
        };                                                                                                   // 1342
                                                                                                             // 1343
        record.proofs.forEach(function (proof) {                                                             // 1344
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1345
          // this would filter.)                                                                             // 1346
          for (var field in proof) {                                                                         // 1347
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1348
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1349
            // return any objects anyway).                                                                   // 1350
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1351
              delete proof[field];                                                                           // 1352
            }                                                                                                // 1353
          }                                                                                                  // 1354
                                                                                                             // 1355
          // Indicate not verified.                                                                          // 1356
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1357
          //   Keybase node-based CLI.                                                                       // 1358
          proof.status = "unverified";                                                                       // 1359
        });                                                                                                  // 1360
                                                                                                             // 1361
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1362
      } else {                                                                                               // 1363
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1364
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1365
        // from Keybase.                                                                                     // 1366
        //                                                                                                   // 1367
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1368
        //   but at present we don't have the ability to verify proofs.                                      // 1369
        KeybaseProfiles.update(keyFingerprint,                                                               // 1370
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1371
      }                                                                                                      // 1372
    });                                                                                                      // 1373
  };                                                                                                         // 1374
                                                                                                             // 1375
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1376
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1377
    var db = this.connection.sandstormDb;                                                                    // 1378
                                                                                                             // 1379
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1380
    if (cursor.count() === 0) {                                                                              // 1381
      // Fire off async update.                                                                              // 1382
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1383
    }                                                                                                        // 1384
                                                                                                             // 1385
    return cursor;                                                                                           // 1386
  });                                                                                                        // 1387
                                                                                                             // 1388
  Meteor.publish("appIndex", function (appId) {                                                              // 1389
    check(appId, String);                                                                                    // 1390
    var db = this.connection.sandstormDb;                                                                    // 1391
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1392
    return cursor;                                                                                           // 1393
  });                                                                                                        // 1394
                                                                                                             // 1395
  Meteor.publish("userPackages", function() {                                                                // 1396
    // Users should be able to see packages that are either:                                                 // 1397
    // 1. referenced by one of their userActions                                                             // 1398
    // 2. referenced by one of their grains                                                                  // 1399
    var self = this;                                                                                         // 1400
    var db = this.connection.sandstormDb;                                                                    // 1401
                                                                                                             // 1402
    // Note that package information, once it is in the database, is static. There's no need to              // 1403
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1404
    // to reactively remove a package from the client side when it is removed on the server, or              // 1405
    // when the client stops using it, because the worst case is the client has a small amount               // 1406
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1407
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1408
    // stuff, we push the package info to the client, and then we never update it.                           // 1409
    //                                                                                                       // 1410
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1411
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1412
    var hasPackage = {};                                                                                     // 1413
    var refPackage = function (packageId) {                                                                  // 1414
      // Ignore dev apps.                                                                                    // 1415
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1416
                                                                                                             // 1417
      if (!hasPackage[packageId]) {                                                                          // 1418
        hasPackage[packageId] = true;                                                                        // 1419
        var pkg = db.getPackage(packageId);                                                                  // 1420
        if (pkg) {                                                                                           // 1421
          self.added("packages", packageId, pkg);                                                            // 1422
        } else {                                                                                             // 1423
          console.error(                                                                                     // 1424
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1425
        }                                                                                                    // 1426
      }                                                                                                      // 1427
    };                                                                                                       // 1428
                                                                                                             // 1429
    // package source 1: packages referred to by actions                                                     // 1430
    var actions = db.userActions(this.userId);                                                               // 1431
    var actionsHandle = actions.observe({                                                                    // 1432
      added: function(newAction) {                                                                           // 1433
        refPackage(newAction.packageId);                                                                     // 1434
      },                                                                                                     // 1435
      changed: function(oldAction, newAction) {                                                              // 1436
        refPackage(newAction.packageId);                                                                     // 1437
      }                                                                                                      // 1438
    });                                                                                                      // 1439
                                                                                                             // 1440
    // package source 2: packages referred to by grains directly                                             // 1441
    var grains = db.userGrains(this.userId);                                                                 // 1442
    var grainsHandle = grains.observe({                                                                      // 1443
      added: function(newGrain) {                                                                            // 1444
        refPackage(newGrain.packageId);                                                                      // 1445
      },                                                                                                     // 1446
      changed: function(oldGrain, newGrain) {                                                                // 1447
        refPackage(newGrain.packageId);                                                                      // 1448
      }                                                                                                      // 1449
    });                                                                                                      // 1450
                                                                                                             // 1451
    this.onStop(function () {                                                                                // 1452
      actionsHandle.stop();                                                                                  // 1453
      grainsHandle.stop();                                                                                   // 1454
    });                                                                                                      // 1455
    this.ready();                                                                                            // 1456
  });                                                                                                        // 1457
}                                                                                                            // 1458
                                                                                                             // 1459
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  SandstormDb.ensureSubscriberHasIdentity = function(publishHandler, identityId) {                           // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber             // 22
    // has a given identity linked. Automatically stops the subscription if the user loses the               // 23
    // identity. Returns a boolean indicating whether the user initially has the identity.                   // 24
                                                                                                             // 25
    var userId = publishHandler.userId;                                                                      // 26
    if (userId === identityId) {                                                                             // 27
      return true;                                                                                           // 28
    } else {                                                                                                 // 29
      var hasIdentityCursor =                                                                                // 30
          Meteor.users.find({$or: [{_id: userId, "loginIdentities.id": identityId},                          // 31
                                   {_id: userId, "nonloginIdentities.id": identityId}]});                    // 32
      if (hasIdentityCursor.count() == 0) {                                                                  // 33
        publishHandler.stop();                                                                               // 34
        return false;                                                                                        // 35
      }                                                                                                      // 36
      var handle = hasIdentityCursor.observe({removed: function () { publishHandler.stop(); }});             // 37
      publishHandler.onStop(function () { handle.stop(); });                                                 // 38
      return true;                                                                                           // 39
    }                                                                                                        // 40
  }                                                                                                          // 41
                                                                                                             // 42
  Meteor.publish("identityProfile", function (identityId) {                                                  // 43
    check(identityId, String);                                                                               // 44
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                  // 45
                                                                                                             // 46
    return Meteor.users.find({_id: identityId},                                                              // 47
      {fields: {                                                                                             // 48
        "profile":1,                                                                                         // 49
        "unverifiedEmail":1,                                                                                 // 50
        "expires": 1,                                                                                        // 51
                                                                                                             // 52
        "services.dev.name":1,                                                                               // 53
                                                                                                             // 54
        "services.google.id":1,                                                                              // 55
        "services.google.email":1,                                                                           // 56
        "services.google.verified_email":1,                                                                  // 57
        "services.google.name":1,                                                                            // 58
        "services.google.picture":1,                                                                         // 59
        "services.google.gender":1,                                                                          // 60
                                                                                                             // 61
        "services.github.id":1,                                                                              // 62
        "services.github.email":1,                                                                           // 63
        "services.github.emails":1,                                                                          // 64
        "services.github.username":1,                                                                        // 65
                                                                                                             // 66
        "services.email.email":1,                                                                            // 67
      }});                                                                                                   // 68
  }),                                                                                                        // 69
                                                                                                             // 70
  Meteor.publish("accountIdentities", function () {                                                          // 71
    // Maybe this should be folded into the "credentials" subscription?                                      // 72
                                                                                                             // 73
    if (!this.userId) return [];                                                                             // 74
                                                                                                             // 75
    return Meteor.users.find(                                                                                // 76
      {_id: this.userId},                                                                                    // 77
      {fields: {                                                                                             // 78
        "profile": 1,                                                                                        // 79
        "verifiedEmail": 1,                                                                                  // 80
        "loginIdentities": 1,                                                                                // 81
        "nonloginIdentities": 1,                                                                             // 82
        "expires": 1,                                                                                        // 83
        "primaryEmail": 1,                                                                                   // 84
      }});                                                                                                   // 85
  });                                                                                                        // 86
                                                                                                             // 87
  makeIdenticon = function (id) {                                                                            // 88
    // We only make identicons client-side.                                                                  // 89
    return undefined;                                                                                        // 90
  }                                                                                                          // 91
                                                                                                             // 92
  var Url = Npm.require("url");                                                                              // 93
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 94
} else {                                                                                                     // 95
  var identiconCache = {};                                                                                   // 96
                                                                                                             // 97
  makeIdenticon = function (id) {                                                                            // 98
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 99
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 100
                                                                                                             // 101
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 102
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 103
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 104
    // anyway to be safe.                                                                                    // 105
    id = id.slice(0, 32);                                                                                    // 106
                                                                                                             // 107
    if (id in identiconCache) {                                                                              // 108
      return identiconCache[id];                                                                             // 109
    }                                                                                                        // 110
                                                                                                             // 111
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 112
    var data = new Identicon(id, 512).toString();                                                            // 113
    var result = "data:image/png;base64," + data;                                                            // 114
    identiconCache[id] = result;                                                                             // 115
    return result;                                                                                           // 116
  }                                                                                                          // 117
                                                                                                             // 118
  httpProtocol = window.location.protocol;                                                                   // 119
}                                                                                                            // 120
                                                                                                             // 121
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 122
                                                                                                             // 123
function staticAssetUrl(id, staticHost) {                                                                    // 124
  if (id) {                                                                                                  // 125
    return staticHost + "/" + id;                                                                            // 126
  } else {                                                                                                   // 127
    return undefined;                                                                                        // 128
  }                                                                                                          // 129
}                                                                                                            // 130
                                                                                                             // 131
function filterHandle(handle) {                                                                              // 132
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 133
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 134
}                                                                                                            // 135
                                                                                                             // 136
function emailToHandle(email) {                                                                              // 137
  // Turn an email address into a handle.                                                                    // 138
                                                                                                             // 139
  if (!email) return undefined;                                                                              // 140
                                                                                                             // 141
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 142
  // on the same account.                                                                                    // 143
  var parts = email.split("@");                                                                              // 144
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 145
                                                                                                             // 146
  var domain = (parts[1]||"").split(".");                                                                    // 147
  if (domain[domain.length - 1] === "name") {                                                                // 148
    // Oh, a .name domain. Let's use                                                                         // 149
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 150
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 151
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 152
                         base)) {                                                                            // 153
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 154
    base = filterHandle(domain[0]);                                                                          // 155
  }                                                                                                          // 156
                                                                                                             // 157
  return filterHandle(base);                                                                                 // 158
}                                                                                                            // 159
                                                                                                             // 160
SandstormDb.fillInProfileDefaults = function(user) {                                                         // 161
  var profile = user.profile;                                                                                // 162
  if (profile.service === "github") {                                                                        // 163
    profile.name = profile.name || user.services.github.username || "Name Unknown";                          // 164
    profile.handle = profile.handle || filterHandle(user.services.github.username) ||                        // 165
        filterHandle(profile.name);                                                                          // 166
  } else if (profile.service === "google") {                                                                 // 167
    profile.name = profile.name || user.services.google.name || "Name Unknown";                              // 168
    profile.handle = profile.handle || emailToHandle(user.services.google.email) ||                          // 169
        filterHandle(profile.name);                                                                          // 170
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                  // 171
  } else if (profile.service === "email") {                                                                  // 172
    var email = user.services.email.email                                                                    // 173
    profile.name = profile.name || email.split("@")[0];                                                      // 174
    profile.handle = profile.handle || emailToHandle(email);                                                 // 175
  } else if (profile.service === "dev") {                                                                    // 176
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                  // 177
    profile.name = profile.name || user.services.dev.name;                                                   // 178
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                          // 179
    profile.pronoun = profile.pronoun ||                                                                     // 180
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 181
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 182
  } else if (profile.service === "demo") {                                                                   // 183
    profile.name = profile.name || "Demo User";                                                              // 184
    profile.handle = profile.handle || "demo";                                                               // 185
  } else {                                                                                                   // 186
    throw new Error("unrecognized identity service: ", profile.service);                                     // 187
  }                                                                                                          // 188
                                                                                                             // 189
  profile.pronoun = profile.pronoun || "neutral";                                                            // 190
}                                                                                                            // 191
                                                                                                             // 192
SandstormDb.fillInIntrinsicName = function(user) {                                                           // 193
  var profile = user.profile;                                                                                // 194
  if (profile.service === "github") {                                                                        // 195
    profile.intrinsicName = user.services.github.username;                                                   // 196
  } else if (profile.service === "google") {                                                                 // 197
    profile.intrinsicName = user.services.google.name;                                                       // 198
    user.privateIntrinsicName = user.services.google.email;                                                  // 199
  } else if (profile.service === "email") {                                                                  // 200
    profile.intrinsicName = user.services.email.email                                                        // 201
  } else if (profile.service === "dev") {                                                                    // 202
    profile.intrinsicName = user.services.dev.name;                                                          // 203
  } else if (profile.service === "demo") {                                                                   // 204
    // No intrinsic name.                                                                                    // 205
  } else {                                                                                                   // 206
    throw new Error("unrecognized identity service: ", profile.service);                                     // 207
  }                                                                                                          // 208
}                                                                                                            // 209
                                                                                                             // 210
SandstormDb.getVerifiedEmails = function(identity) {                                                         // 211
  if (identity.services.google && identity.services.google.email &&                                          // 212
      identity.services.google.verified_email) {                                                             // 213
    return [identity.services.google.email];                                                                 // 214
  } else if (identity.services.email) {                                                                      // 215
    return [identity.services.email.email];                                                                  // 216
  } else if (identity.services.github && identity.services.github.emails) {                                  // 217
    return _.pluck(_.filter(identity.services.github.emails,                                                 // 218
                            function(email) { return email.verified; }),                                     // 219
                   "email");                                                                                 // 220
  }                                                                                                          // 221
  return [];                                                                                                 // 222
}                                                                                                            // 223
                                                                                                             // 224
SandstormDb.fillInPictureUrl = function(user) {                                                              // 225
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 226
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) ||                              // 227
    makeIdenticon(user._id);                                                                                 // 228
}                                                                                                            // 229
                                                                                                             // 230
SandstormDb.getUserIdentityIds = function (user) {                                                           // 231
  // Given an account user object, returns an array containing the ID of each identity linked to the         // 232
  // account. Always returns the most recently added login identity first.                                   // 233
  if (user && user.loginIdentities) {                                                                        // 234
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                    // 235
  } else {                                                                                                   // 236
    return [];                                                                                               // 237
  }                                                                                                          // 238
};                                                                                                           // 239
                                                                                                             // 240
SandstormDb.getUserEmails = function (user) {                                                                // 241
  // Given a user object, returns an array containing all email addresses associated with that user.         // 242
  // Each entry in the array is an object of the form:                                                       // 243
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 244
  //                                                                                                         // 245
  // At most one entry in the result has `primary = true`.                                                   // 246
  //                                                                                                         // 247
  // TODO(cleanup): This actually does need to query the database to fetch profile information               // 248
  //   for linked identities, so it probably makes more sense for it to be a non-static method               // 249
  //   on SandstormDb.                                                                                       // 250
                                                                                                             // 251
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                    // 252
  verifiedEmails = {};                                                                                       // 253
  unverifiedEmails = {};                                                                                     // 254
                                                                                                             // 255
  identityIds.forEach(function (id) {                                                                        // 256
    var identity = Meteor.users.findOne({_id: id});                                                          // 257
    if (identity && identity.services) {                                                                     // 258
      SandstormDb.getVerifiedEmails(identity).forEach(function(verifiedEmail) {                              // 259
        if (verifiedEmail) {                                                                                 // 260
          verifiedEmails[verifiedEmail] = true;                                                              // 261
        }                                                                                                    // 262
      });                                                                                                    // 263
    }                                                                                                        // 264
    if (identity && identity.unverifiedEmail) {                                                              // 265
      unverifiedEmails[identity.unverifiedEmail] = true;                                                     // 266
    }                                                                                                        // 267
  });                                                                                                        // 268
                                                                                                             // 269
  var result = [];                                                                                           // 270
  _.keys(verifiedEmails).map(function (email) {                                                              // 271
    result.push({email: email,                                                                               // 272
                 verified: true,                                                                             // 273
                 primary: email === user.primaryEmail});                                                     // 274
  });                                                                                                        // 275
  _.keys(unverifiedEmails).map(function (email) {                                                            // 276
    if (!(email in verifiedEmails)) { result.push({email: email, verified: false}) }                         // 277
  });                                                                                                        // 278
                                                                                                             // 279
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                     // 280
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                         // 281
    result[0].primary = true;                                                                                // 282
  }                                                                                                          // 283
  return result;                                                                                             // 284
}                                                                                                            // 285
                                                                                                             // 286
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  if (user.loginIdentities) {                                                                                // 72
    // it's an account                                                                                       // 73
    check(user, {_id: String,                                                                                // 74
                 createdAt: Date,                                                                            // 75
                 isAdmin: Match.Optional(Boolean),                                                           // 76
                 hasCompletedSignup: Match.Optional(Boolean),                                                // 77
                 signupKey: Match.Optional(String),                                                          // 78
                 signupNote: Match.Optional(String),                                                         // 79
                 signupEmail: Match.Optional(String),                                                        // 80
                 expires: Match.Optional(Date),                                                              // 81
                 loginIdentities: [{id: String}],                                                            // 82
                 nonloginIdentities: [{id: String}]});                                                       // 83
    return user;                                                                                             // 84
  }                                                                                                          // 85
                                                                                                             // 86
  // Check profile.                                                                                          // 87
  if (options.profile) {                                                                                     // 88
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 89
    check(options.profile, Match.ObjectIncluding({                                                           // 90
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 91
      handle: Match.Optional(ValidHandle),                                                                   // 92
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 93
    }));                                                                                                     // 94
  }                                                                                                          // 95
                                                                                                             // 96
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 97
                                                                                                             // 98
  if (options.unverifiedEmail) {                                                                             // 99
    user.unverifiedEmail = options.unverifiedEmail;                                                          // 100
  }                                                                                                          // 101
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                // 102
                                                                                                             // 103
  // Try downloading avatar.                                                                                 // 104
  var url = userPictureUrl(user);                                                                            // 105
  if (url) {                                                                                                 // 106
    var assetId = fetchPicture(url);                                                                         // 107
    if (assetId) {                                                                                           // 108
      user.profile.picture = assetId;                                                                        // 109
    }                                                                                                        // 110
  }                                                                                                          // 111
                                                                                                             // 112
  var serviceUserId;                                                                                         // 113
  if (user.services && user.services.dev) {                                                                  // 114
    check(user.services.dev, {name: String, isAdmin: Boolean, hasCompletedSignup: Boolean});                 // 115
    serviceUserId = user.services.dev.name;                                                                  // 116
    user.profile.service = "dev";                                                                            // 117
  } else if ("expires" in user) {                                                                            // 118
    serviceUserId = user._id;                                                                                // 119
    user.profile.service = "demo";                                                                           // 120
  } else if (user.services && user.services.email) {                                                         // 121
    check(user.services.email,                                                                               // 122
          {email: String,                                                                                    // 123
           tokens: [{digest: String, algorithm: String, createdAt: Date}]});                                 // 124
    serviceUserId = user.services.email.email;                                                               // 125
    user.profile.service = "email";                                                                          // 126
  } else if (user.services && "google" in user.services) {                                                   // 127
    serviceUserId = user.services.google.id;                                                                 // 128
    user.profile.service = "google";                                                                         // 129
  } else if (user.services && "github" in user.services) {                                                   // 130
    serviceUserId = user.services.github.id;                                                                 // 131
    user.profile.service = "github";                                                                         // 132
  } else {                                                                                                   // 133
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " +                      // 134
                           JSON.stringify(user));                                                            // 135
  }                                                                                                          // 136
  user._id = Crypto.createHash("sha256")                                                                     // 137
    .update(user.profile.service + ":" + serviceUserId).digest("hex");                                       // 138
                                                                                                             // 139
  return user;                                                                                               // 140
});                                                                                                          // 141
                                                                                                             // 142
// TODO delete this obsolete index.                                                                          // 143
Meteor.users._ensureIndex("identities.id", {unique: 1, sparse: 1});                                          // 144
                                                                                                             // 145
Meteor.users._ensureIndex("loginIdentities.id", {unique: 1, sparse: 1});                                     // 146
Meteor.users._ensureIndex("nonloginIdentities.id", {sparse: 1});                                             // 147
                                                                                                             // 148
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
                                                                                                             // 8
var updateLoginStyleToRedirect = function() {                                                                // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 10
  ["google", "github"].forEach(function(serviceName) {                                                       // 11
    var config = configurations.findOne({service: serviceName});                                             // 12
    if (config && config.loginStyle !== "redirect") {                                                        // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 14
    }                                                                                                        // 15
  });                                                                                                        // 16
};                                                                                                           // 17
                                                                                                             // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 19
  // In the before time, Google and Github login were enabled by default.                                    // 20
  //                                                                                                         // 21
  // This actually didn't make much sense, required the first user to configure                              // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 23
  // secure.                                                                                                 // 24
  //                                                                                                         // 25
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 26
  // (since they need to be configured anyway) but since users may have never                                // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 28
  // we can't just use the value in the Settings collection, since it might                                  // 29
  // never have been set.                                                                                    // 30
  //                                                                                                         // 31
  // Thus, if the service is configured but absent from Settings, we should                                  // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 33
  // depend on what value is in Settings and default to false without breaking                               // 34
  // user installations.                                                                                     // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 36
  ["google", "github"].forEach(function(serviceName) {                                                       // 37
    var config = configurations.findOne({service: serviceName});                                             // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 39
    if (config && !serviceConfig) {                                                                          // 40
      // Only explicitly enable the login service if:                                                        // 41
      // 1) the service is already configured                                                                // 42
      // 2) there is no sandstorm configuration already present (the user was                                // 43
      //    using the previous default behavior).                                                            // 44
      Settings.insert({_id: serviceName, value: true});                                                      // 45
    }                                                                                                        // 46
  });                                                                                                        // 47
};                                                                                                           // 48
                                                                                                             // 49
var denormalizeInviteInfo = function() {                                                                     // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 52
  // which has more information about the key, namely a freeform note entered by the admin when              // 53
  // they created the key. In the case that the email invite form was used, the note has the form            // 54
  // "E-mail invite to <address>".                                                                           // 55
  //                                                                                                         // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 57
  // information about the invite terms. Namely, for email invites we want to show the address               // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 60
  // from the SignupKeys table.                                                                              // 61
                                                                                                             // 62
  Meteor.users.find().forEach(function (user) {                                                              // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 65
      if (signupInfo && signupInfo.note) {                                                                   // 66
        var newFields = { signupNote: signupInfo.note };                                                     // 67
                                                                                                             // 68
        var prefix = "E-mail invite to ";                                                                    // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 71
        }                                                                                                    // 72
                                                                                                             // 73
        Meteor.users.update(user._id, {$set: newFields});                                                    // 74
      }                                                                                                      // 75
    }                                                                                                        // 76
  });                                                                                                        // 77
}                                                                                                            // 78
                                                                                                             // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 81
    ApiTokens.insert({                                                                                       // 82
      grainId: roleAssignment.grainId,                                                                       // 83
      userId: roleAssignment.sharer,                                                                         // 84
      roleAssignment: roleAssignment.roleAssignment,                                                         // 85
      petname: roleAssignment.petname,                                                                       // 86
      created: roleAssignment.created,                                                                       // 87
      owner: {user: {userId: roleAssignment.recipient,                                                       // 88
                     title: roleAssignment.title}},                                                          // 89
    });                                                                                                      // 90
  });                                                                                                        // 91
}                                                                                                            // 92
                                                                                                             // 93
function fixOasisStorageUsageStats() {}                                                                      // 94
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 95
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 96
// with a no-op.                                                                                             // 97
                                                                                                             // 98
function fetchProfilePictures() {                                                                            // 99
  Meteor.users.find({}).forEach(function (user) {                                                            // 100
    var url = userPictureUrl(user);                                                                          // 101
    if (url) {                                                                                               // 102
      console.log("Fetching user picture:", url);                                                            // 103
      var assetId = fetchPicture(url);                                                                       // 104
      if (assetId) {                                                                                         // 105
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 106
      }                                                                                                      // 107
    }                                                                                                        // 108
  });                                                                                                        // 109
}                                                                                                            // 110
                                                                                                             // 111
function assignPlans() {                                                                                     // 112
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 113
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 114
  }                                                                                                          // 115
}                                                                                                            // 116
                                                                                                             // 117
function removeKeyrings() {                                                                                  // 118
  // These blobs full of public keys were not intended to find their way into mongo and while                // 119
  // harmless they slow things down because they're huge. Remove them.                                       // 120
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 121
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 122
      {multi: true});                                                                                        // 123
}                                                                                                            // 124
                                                                                                             // 125
function useLocalizedTextInUserActions() {                                                                   // 126
  function toLocalizedText(newObj, oldObj, field) {                                                          // 127
    if (field in oldObj) {                                                                                   // 128
      if (typeof oldObj[field] === "string") {                                                               // 129
        newObj[field] = {defaultText: oldObj[field]};                                                        // 130
      } else {                                                                                               // 131
        newObj[field] = oldObj[field];                                                                       // 132
      }                                                                                                      // 133
    }                                                                                                        // 134
  }                                                                                                          // 135
  UserActions.find({}).forEach(function (userAction) {                                                       // 136
    var fields = {};                                                                                         // 137
    toLocalizedText(fields, userAction, "appTitle");                                                         // 138
    toLocalizedText(fields, userAction, "title");                                                            // 139
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 140
    UserActions.update(userAction._id, {$set: fields});                                                      // 141
  });                                                                                                        // 142
}                                                                                                            // 143
                                                                                                             // 144
function verifyAllPgpSignatures() {                                                                          // 145
  Packages.find({}).forEach(function (pkg) {                                                                 // 146
    try {                                                                                                    // 147
      console.log("checking PGP signature for package:", pkg._id);                                           // 148
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 149
      if (info.authorPgpKeyFingerprint) {                                                                    // 150
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 151
        Packages.update(pkg._id,                                                                             // 152
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 153
      } else {                                                                                               // 154
        console.log("  no signature");                                                                       // 155
      }                                                                                                      // 156
    } catch (err) {                                                                                          // 157
      console.error(err.stack);                                                                              // 158
    }                                                                                                        // 159
  });                                                                                                        // 160
}                                                                                                            // 161
                                                                                                             // 162
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 163
  var Crypto = Npm.require("crypto");                                                                        // 164
  Meteor.users.find().forEach(function (user) {                                                              // 165
    var identity = {};                                                                                       // 166
    var serviceUserId;                                                                                       // 167
    if ("devName" in user) {                                                                                 // 168
      identity.service = "dev";                                                                              // 169
      serviceUserId = user.devName;                                                                          // 170
    } else if ("expires" in user) {                                                                          // 171
      identity.service = "demo";                                                                             // 172
      serviceUserId = user._id;                                                                              // 173
    } else if (user.services && "google" in user.services) {                                                 // 174
      identity.service = "google";                                                                           // 175
      if (user.services.google.email && user.services.google.verified_email) {                               // 176
        identity.verifiedEmail = user.services.google.email;                                                 // 177
      }                                                                                                      // 178
      serviceUserId = user.services.google.id;                                                               // 179
    } else if (user.services && "github" in user.services) {                                                 // 180
      identity.service = "github";                                                                           // 181
      identity.unverifiedEmail = user.services.github.email;                                                 // 182
      serviceUserId = user.services.github.id;                                                               // 183
    } else if (user.services && "emailToken" in user.services) {                                             // 184
      identity.service = "emailToken";                                                                       // 185
      identity.verifiedEmail = user.services.emailToken.email;                                               // 186
      serviceUserId = user.services.emailToken.email;                                                        // 187
    }                                                                                                        // 188
                                                                                                             // 189
    identity.id = Crypto.createHash("sha256")                                                                // 190
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 191
                                                                                                             // 192
    if (user.profile) {                                                                                      // 193
      if (user.profile.name) {                                                                               // 194
        identity.name = user.profile.name;                                                                   // 195
      }                                                                                                      // 196
      if (user.profile.handle) {                                                                             // 197
        identity.handle = user.profile.handle;                                                               // 198
      }                                                                                                      // 199
      if (user.profile.picture) {                                                                            // 200
        identity.picture = user.profile.picture;                                                             // 201
      }                                                                                                      // 202
      if (user.profile.pronoun) {                                                                            // 203
        identity.pronoun = user.profile.pronoun;                                                             // 204
      }                                                                                                      // 205
      if (user.profile.email) {                                                                              // 206
        identity.unverifiedEmail = user.profile.email;                                                       // 207
      }                                                                                                      // 208
    }                                                                                                        // 209
    identity.main = true;                                                                                    // 210
                                                                                                             // 211
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 212
                                                                                                             // 213
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 214
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 215
    ApiTokens.update({userId: user._id},                                                                     // 216
                     {$set: {identityId: identity.id}},                                                      // 217
                     {multi: true});                                                                         // 218
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 219
                     {$set: {"owner.user.identityId": identity.id}},                                         // 220
                     {multi: true});                                                                         // 221
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 222
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 223
                     {multi: true});                                                                         // 224
                                                                                                             // 225
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 226
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 227
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 228
                            {multi: true}) > 0);                                                             // 229
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 230
    // there may be many matches, we need to repeat until no documents are modified.                         // 231
                                                                                                             // 232
  });                                                                                                        // 233
                                                                                                             // 234
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 235
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 236
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 237
  // form of API token.                                                                                      // 238
}                                                                                                            // 239
                                                                                                             // 240
function appUpdateSettings() {                                                                               // 241
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 242
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 243
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 244
}                                                                                                            // 245
                                                                                                             // 246
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 247
  Meteor.users.find().forEach(function (user) {                                                              // 248
    if (user.identities.length != 1) {                                                                       // 249
      throw new Error("User does not have exactly one identity: ", user);                                    // 250
    }                                                                                                        // 251
    var identity = user.identities[0];                                                                       // 252
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 253
                                                                                                             // 254
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 255
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 256
                                                                                                             // 257
    var serviceObject = {};                                                                                  // 258
    var fieldsToUnset = {}                                                                                   // 259
                                                                                                             // 260
    if (identity.service === "dev") {                                                                        // 261
      serviceObject.name = user.devName;                                                                     // 262
      fieldsToUnset.devName = 1;                                                                             // 263
    } else if (identity.service === "emailToken") {                                                          // 264
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 265
      serviceObject.email = user.services.emailToken.email;                                                  // 266
      fieldsToUnset["services.emailToken"] = 1;                                                              // 267
    }                                                                                                        // 268
                                                                                                             // 269
    newIdentity.service = {};                                                                                // 270
    newIdentity.service[identity.service] = serviceObject;                                                   // 271
                                                                                                             // 272
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 273
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 274
      modifier["$unset"] = fieldsToUnset;                                                                    // 275
    }                                                                                                        // 276
                                                                                                             // 277
    Meteor.users.update({_id: user._id}, modifier);                                                          // 278
  });                                                                                                        // 279
}                                                                                                            // 280
                                                                                                             // 281
function repairEmailIdentityIds() {                                                                          // 282
  var Crypto = Npm.require("crypto");                                                                        // 283
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 284
    if (user.identities.length != 1) {                                                                       // 285
      throw new Error("User does not have exactly one identity: ", user);                                    // 286
    }                                                                                                        // 287
    var identity = user.identities[0];                                                                       // 288
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 289
                             "profile");                                                                     // 290
    newIdentity.service = {email: identity.service.emailToken};                                              // 291
    newIdentity.id = Crypto.createHash("sha256")                                                             // 292
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 293
                                                                                                             // 294
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 295
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 296
    ApiTokens.update({identityId: identity.id},                                                              // 297
                     {$set: {identityId: newIdentity.id}},                                                   // 298
                     {multi: true});                                                                         // 299
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 300
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 301
                     {multi: true});                                                                         // 302
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 303
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 304
                     {multi: true});                                                                         // 305
                                                                                                             // 306
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 307
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 308
                            {multi: true}) > 0);                                                             // 309
                                                                                                             // 310
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 311
  });                                                                                                        // 312
}                                                                                                            // 313
                                                                                                             // 314
function splitAccountUsersAndIdentityUsers() {                                                               // 315
  Meteor.users.find({"identities": {$exists: true}}).forEach(function (user) {                               // 316
    if (user.identities.length != 1) {                                                                       // 317
      throw new Error("User does not have exactly one identity: ", user);                                    // 318
    }                                                                                                        // 319
    var identity = user.identities[0];                                                                       // 320
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires")                                    // 321
    identityUser._id = identity.id;                                                                          // 322
    identityUser.profile = identity.profile;                                                                 // 323
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                             // 324
    identityUser.profile.service = Object.keys(identity.service)[0];                                         // 325
                                                                                                             // 326
    // Updating this user needs to be a two step process because the `services` field typically              // 327
    // contains subfields that are constrained to be unique by Mongo indices.                                // 328
    identityUser.stagedServices = _.omit(user.services, "resume");                                           // 329
    if (identity.service.dev) {                                                                              // 330
      identityUser.stagedServices.dev = identity.service.dev;                                                // 331
    } else if (identity.service.email) {                                                                     // 332
      identityUser.stagedServices.email = identity.service.email;                                            // 333
    }                                                                                                        // 334
                                                                                                             // 335
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires",                              // 336
                             "isAdmin", "signupKey", "signupNote", "signupEmail",                            // 337
                             "plan", "storageUsage", "isAppDemoUser", "appDemoId",                           // 338
                             "payments", "dailySentMailCount", "hasCompletedSignup");                        // 339
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                  // 340
    accountUser.nonloginIdentities = [];                                                                     // 341
    if (user.services && user.services.resume) {                                                             // 342
      accountUser.services = {resume: user.services.resume};                                                 // 343
    }                                                                                                        // 344
                                                                                                             // 345
    accountUser.stashedOldUser = user;                                                                       // 346
                                                                                                             // 347
    ApiTokens.update({identityId: identityUser._id}, {$set: {accountId: user._id}},                          // 348
                     {multi: true});                                                                         // 349
                                                                                                             // 350
    Meteor.users.upsert({_id: identityUser._id}, identityUser);                                              // 351
    Meteor.users.update({_id: user._id}, accountUser);                                                       // 352
  });                                                                                                        // 353
                                                                                                             // 354
  Meteor.users.find({"stagedServices": {$exists: true}}).forEach(function (identity) {                       // 355
    Meteor.users.update({_id: identity._id}, {$unset: {stagedServices: 1},                                   // 356
                                              $set: {services: identity.stagedServices}});                   // 357
  });                                                                                                        // 358
}                                                                                                            // 359
                                                                                                             // 360
// This must come after all the functions named within are defined.                                          // 361
// Only append to this list!  Do not modify or remove list entries;                                          // 362
// doing so is likely change the meaning and semantics of user databases.                                    // 363
var MIGRATIONS = [                                                                                           // 364
  updateLoginStyleToRedirect,                                                                                // 365
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 366
  denormalizeInviteInfo,                                                                                     // 367
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 368
  fixOasisStorageUsageStats,                                                                                 // 369
  fetchProfilePictures,                                                                                      // 370
  assignPlans,                                                                                               // 371
  removeKeyrings,                                                                                            // 372
  useLocalizedTextInUserActions,                                                                             // 373
  verifyAllPgpSignatures,                                                                                    // 374
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 375
  appUpdateSettings,                                                                                         // 376
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 377
  repairEmailIdentityIds,                                                                                    // 378
  splitAccountUsersAndIdentityUsers,                                                                         // 379
];                                                                                                           // 380
                                                                                                             // 381
function migrateToLatest() {                                                                                 // 382
  if (Meteor.settings.replicaNumber) {                                                                       // 383
    // This is a replica. Wait for the first replica to perform migrations.                                  // 384
                                                                                                             // 385
    console.log("Waiting for migrations on replica zero...");                                                // 386
                                                                                                             // 387
    var done = new Future();                                                                                 // 388
    var change = function (doc) {                                                                            // 389
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 390
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 391
    }                                                                                                        // 392
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 393
      added: change,                                                                                         // 394
      changed: change                                                                                        // 395
    });                                                                                                      // 396
                                                                                                             // 397
    done.wait();                                                                                             // 398
    observer.stop();                                                                                         // 399
    console.log("Migrations have completed on replica zero.");                                               // 400
                                                                                                             // 401
  } else {                                                                                                   // 402
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 403
    var start;                                                                                               // 404
    if (!applied) {                                                                                          // 405
      // Migrations table is not yet seeded with a value.  This means it has                                 // 406
      // applied 0 migrations.  Persist this.                                                                // 407
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 408
      start = 0;                                                                                             // 409
    } else {                                                                                                 // 410
      start = applied.value;                                                                                 // 411
    }                                                                                                        // 412
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 413
                                                                                                             // 414
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 415
      // Apply migration i, then record that migration i was successfully run.                               // 416
      console.log("Applying migration " + (i+1));                                                            // 417
      MIGRATIONS[i]();                                                                                       // 418
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 419
      console.log("Applied migration " + (i+1));                                                             // 420
    }                                                                                                        // 421
  }                                                                                                          // 422
}                                                                                                            // 423
                                                                                                             // 424
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 425
                                                                                                             // 426
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
